'use strict';

const stripOptionsFromArgs = (args) => {
    const opts = args[args.length - 1];
    if (typeof opts === 'object' && opts.constructor === Object) {
        args.splice(args.length - 1, 1);
        return opts;
    }
    return {};
};
const source = (re) => {
    if (!re)
        return null;
    if (typeof re === 'string')
        return re;
    return re.source;
};
const concat = (...args) => args.map((x) => source(x)).join('');
const lookahead = (re) => concat('(?=', re, ')');
const optional = (re) => concat('(?:', re, ')?');
const either = (...args) => {
    const opts = stripOptionsFromArgs(args);
    return ([
        '(',
        opts.capture ? '' : '?:',
        args.map((x) => source(x)).join('|'),
        ')',
    ].join(''));
};
const countMatchGroups = (re) => (new RegExp(`${re.toString()}|`).exec('').length - 1);
/**
 * Does lexeme start with a regular expression match at the beginning.
 *
 * @param {RegExp} re
 * @param {string} lexeme
 */
const startsWith = (re, lexeme) => {
    const match = re && re.exec(lexeme);
    return match && match.index === 0;
};
const BACKREF_RE = /\[(?:[^\\\]]|\\.)*\]|\(\??|\\([1-9][0-9]*)|\\./;
const _rewriteBackreferences = (regexps, { joinWith }) => {
    let numCaptures = 0;
    return regexps.map((regex) => {
        numCaptures += 1;
        const offset = numCaptures;
        let re = source(regex);
        let out = '';
        while (re.length > 0) {
            const match = BACKREF_RE.exec(re);
            if (!match) {
                out += re;
                break;
            }
            out += re.substring(0, match.index);
            re = re.substring(match.index + match[0].length);
            if (match[0][0] === '\\' && match[1]) {
                // Adjust the backreference.
                out += `\\${String(Number(match[1]) + offset)}`;
            }
            else {
                out += match[0];
                if (match[0] === '(') {
                    numCaptures += 1;
                }
            }
        }
        return out;
    }).map((re) => `(${re})`).join(joinWith);
};

const escapeHTML = (value) => (value.replace(/&/g, '&amp;')
    .replace(/</g, '&lt;')
    .replace(/>/g, '&gt;')
    .replace(/"/g, '&quot;')
    .replace(/'/g, '&#x27;'));
const inherit = (original, ...objects) => {
    const result = Object.create(null);
    Object.keys(original).forEach((key) => {
        result[key] = original[key];
    });
    objects.forEach((obj) => {
        Object.keys(obj).forEach((key) => {
            result[key] = obj[key];
        });
    });
    return result;
};
const langRe = (value, global, language) => new RegExp(source(value), [
    'm',
    language.caseInsensitive ? 'i' : '',
    language.unicodeRegex ? 'u' : '',
    global ? 'g' : '',
].join(''));

/**
 * Skip a match if it has a preceding dot
 *
 * This is used for `beginKeywords` to prevent matching expressions such as
 * `bob.keyword.do()`. The mode compiler automatically wires this up as a
 * special _internal_ 'on:begin' callback for modes with `beginKeywords`
 * @param {RegExpMatchArray} match
 * @param {CallbackResponse} response
 */
const skipIfHasPrecedingDot = (match, response) => {
    const before = match.input[match.index - 1];
    if (before === '.') {
        response.ignoreMatch();
    }
};
const scopeClassName = (mode) => {
    if (mode.className !== undefined) {
        mode.scope = mode.className;
        delete mode.className;
    }
};
const beginKeywords = (mode, parent) => {
    if (!parent)
        return undefined;
    if (!mode.beginKeywords)
        return undefined;
    mode.begin = `\\b(${mode.beginKeywords.split(' ').join('|')})(?!\\.)(?=\\b|\\s)`;
    mode.__beforeBegin = skipIfHasPrecedingDot;
    mode.keywords = mode.keywords || mode.beginKeywords;
    delete mode.beginKeywords;
    if (mode.relevance === undefined) {
        mode.relevance = 0;
    }
    return undefined;
};
/**
 * Allow `illegal` to contain an array of illegal values.
 */
const compileIllegal = (mode) => {
    if (!Array.isArray(mode.illegal))
        return undefined;
    mode.illegal = either(...mode.illegal);
    return undefined;
};
/**
 * `match` to match a single expression for readability.
 */
const compileMatch = (mode) => {
    if (!mode.match)
        return undefined;
    if (mode.begin || mode.end) {
        throw new Error('begin & end are not supported with match');
    }
    mode.begin = mode.match;
    delete mode.match;
    return undefined;
};
/**
 * Provides the default 1 relevance to all modes.
 */
const compileRelevance = (mode) => {
    if (mode.relevance === undefined) {
        mode.relevance = 1;
    }
};

const beforeMatchExt = (mode) => {
    if (!mode.beforeMatch)
        return undefined;
    if (mode.starts) {
        throw new Error('beforeMatch cannot be used with starts');
    }
    const originalMode = { ...mode };
    Object.keys(mode).forEach((key) => { delete mode[key]; });
    mode.keywords = originalMode.keywords;
    mode.begin = concat(originalMode.beforeMatch, lookahead(originalMode.begin));
    mode.starts = {
        relevance: 0,
        contains: [Object.assign(originalMode, { endsParent: true })],
    };
    mode.relevance = 0;
    delete originalMode.beforeMatch;
    return undefined;
};

const MultiClassError = new Error();
/**
 * Renumbers labeled scope names to account for additional inner match
 * groups that otherwise would break everything.
 *
 * Lets say we 3 match scopes:
 *
 *   { 1 => ..., 2 => ..., 3 => ... }
 *
 * So what we need is a clean match like this:
 *
 *   (a)(b)(c) => [ "a", "b", "c" ]
 *
 * But this falls apart with inner match groups:
 *
 * (a)(((b)))(c) => ["a", "b", "b", "b", "c" ]
 *
 * Our scopes are now "out of alignment" and we're repeating `b` 3 times.
 * What needs to happen is the numbers are remapped:
 *
 *   { 1 => ..., 2 => ..., 5 => ... }
 *
 * We also need to know that the ONLY groups that should be output
 * are 1, 2, and 5.  This function handles this behavior.
 *
 * @param {CompiledMode} mode
 * @param {(string | RegExp)[]} regexes
 * @param {{ key: "beginScope" | "endScope" }} opts
 */
const remapScopeNames = (mode, regexes, { key }) => {
    let offset = 0;
    const scopeNames = mode[key];
    const emit = {};
    const positions = {};
    for (let i = 1; i <= regexes.length; i++) {
        positions[i + offset] = scopeNames[i];
        emit[i + offset] = true;
        offset += countMatchGroups(regexes[i - 1]);
    }
    mode[key] = positions;
    mode[key]._emit = emit;
    mode[key]._multi = true;
};
const beginMultiClass = (mode) => {
    if (!Array.isArray(mode.begin))
        return undefined;
    if (mode.skip || mode.excludeBegin || mode.returnBegin) {
        throw MultiClassError;
    }
    if (typeof mode.beginScope !== 'object' || mode.beginScope === null) {
        throw MultiClassError;
    }
    remapScopeNames(mode, mode.begin, { key: 'beginScope' });
    mode.begin = _rewriteBackreferences(mode.begin, { joinWith: '' });
    return undefined;
};
const endMultiClass = (mode) => {
    if (!Array.isArray(mode.end))
        return undefined;
    if (mode.skip || mode.excludeEnd || mode.returnEnd) {
        throw MultiClassError;
    }
    if (typeof mode.endScope !== 'object' || mode.endScope === null) {
        throw MultiClassError;
    }
    remapScopeNames(mode, mode.end, { key: 'endScope' });
    mode.end = _rewriteBackreferences(mode.end, { joinWith: '' });
    return undefined;
};
/**
 * this exists only to allow `scope: {}` to be used beside `match:`
 * Otherwise `beginScope` would necessary and that would look weird

  {
    match: [ /def/, /\w+/ ]
    scope: { 1: "keyword" , 2: "title" }
  }

 * @param {CompiledMode} mode
 */
const scopeSugar = (mode) => {
    if (mode.scope && typeof mode.scope === 'object' && mode.scope !== null) {
        mode.beginScope = mode.scope;
        delete mode.scope;
    }
};
const multiClass = (mode) => {
    scopeSugar(mode);
    if (typeof mode.beginScope === 'string') {
        mode.beginScope = { _wrap: mode.beginScope };
    }
    if (typeof mode.endScope === 'string') {
        mode.endScope = { _wrap: mode.endScope };
    }
    beginMultiClass(mode);
    endMultiClass(mode);
};

// keywords that should have no default relevance value
const COMMON_KEYWORDS = [
    'of',
    'and',
    'for',
    'in',
    'not',
    'or',
    'if',
    'then',
    'parent',
    'list',
    'value', // common variable name
];
const DEFAULT_KEYWORD_SCOPE = 'keyword';
/**
 * Determines if a given keyword is common or not.
 *
 * @param {string} keyword
 */
const commonKeyword = (keyword) => COMMON_KEYWORDS.includes(keyword.toLowerCase());
/**
 * Return the proper score for a given keyword.
 *
 * Also takes into account comment keywords, which will be scored 0 UNLESS
 * another score has been manually assigned.
 *
 * @param {string} keyword
 * @param {string} providedScore
 */
const scoreForKeyword = (keyword, providedScore) => {
    if (providedScore) {
        return Number(providedScore);
    }
    return commonKeyword(keyword) ? 0 : 1;
};
/**
 * Given raw keywords from a language definition, compile them.
 *
 * @param {RawKeywords} rawKeywords
 * @param {boolean} caseInsensitive
 * @param {string} scopeName
 *
 * @returns {KeywordDict} compiled keywords.
 */
const compileKeywords = (rawKeywords, caseInsensitive, scopeName = DEFAULT_KEYWORD_SCOPE) => {
    const compiledKeywords = Object.create(null);
    // utility functions
    /**
     * Compiles an individual list of keywords
     *
     * Ex: "for if when while|5"
     *
     * @param {string} _scopeName
     * @param {string[]} keywordList
     */
    const compileList = (_scopeName, keywordList) => {
        let _keywordList = keywordList;
        if (caseInsensitive) {
            _keywordList = keywordList.map((x) => x.toLowerCase());
        }
        _keywordList.forEach((keyword) => {
            const pair = keyword.split('|');
            compiledKeywords[pair[0]] = [_scopeName, scoreForKeyword(pair[0], pair[1])];
        });
    };
    if (typeof rawKeywords === 'string') {
        compileList(scopeName, rawKeywords.split(' '));
    }
    else if (Array.isArray(rawKeywords)) {
        compileList(scopeName, rawKeywords);
    }
    else {
        Object.keys(rawKeywords).forEach((_scopeName) => {
            // collapse all our objects back into the parent object
            Object.assign(compiledKeywords, compileKeywords(rawKeywords[_scopeName], caseInsensitive, _scopeName));
        });
    }
    return compiledKeywords;
};

/**
 Stores multiple regular expressions and allows you to quickly search for
 them all in a string simultaneously - returning the first match.  It does
 this by creating a huge (a|b|c) regex - each individual item wrapped with ()
 and joined by `|` - using match groups to track position.  When a match is
 found checking which position in the array has content allows us to figure
 out which of the original regexes / match groups triggered the match.

 The match object itself (the result of `Regex.exec`) is returned but also
 enhanced by merging in any meta-data that was registered with the regex.
 This is how we keep track of which mode matched, and what type of rule
 (`illegal`, `begin`, end, etc).
*/
class MultiRegex {
    _language;
    matchIndexes;
    regexes;
    matchAt;
    position;
    matcherRe;
    lastIndex;
    constructor(langauge) {
        this._language = langauge;
        this.matchIndexes = {};
        this.regexes = [];
        this.matchAt = 1;
        this.position = 0;
    }
    addRule = (re, opts) => {
        opts.position = this.position;
        this.position += 1;
        this.matchIndexes[this.matchAt] = opts;
        this.regexes.push([opts, re]);
        this.matchAt += countMatchGroups(re) + 1;
    };
    compile = () => {
        if (this.regexes.length === 0) {
            this.exec = () => null;
        }
        const terminators = this.regexes.map((el) => el[1]);
        this.matcherRe = langRe(_rewriteBackreferences(terminators, { joinWith: '|' }), true, this._language);
        this.lastIndex = 0;
    };
    exec = (str) => {
        this.matcherRe.lastIndex = this.lastIndex;
        const match = this.matcherRe.exec(str);
        if (!match)
            return null;
        const index = match.findIndex((el, i) => i > 0 && el !== undefined);
        const matchData = this.matchIndexes[index];
        match.splice(0, index);
        return Object.assign(match, matchData);
    };
}

/*
  Created to solve the key deficiently with MultiRegex - there is no way to
  test for multiple matches at a single location.  Why would we need to do
  that?  In the future a more dynamic engine will allow certain matches to be
  ignored.  An example: if we matched say the 3rd regex in a large group but
  decided to ignore it - we'd need to started testing again at the 4th
  regex... but MultiRegex itself gives us no real way to do that.

  So what this class creates MultiRegexs on the fly for whatever search
  position they are needed.

  NOTE: These additional MultiRegex objects are created dynamically.  For most
  grammars most of the time we will never actually need anything more than the
  first MultiRegex - so this shouldn't have too much overhead.

  Say this is our search group, and we match regex3, but wish to ignore it.

    regex1 | regex2 | regex3 | regex4 | regex5    ' ie, startAt = 0

  What we need is a new MultiRegex that only includes the remaining
  possibilities:

    regex4 | regex5                               ' ie, startAt = 3

  This class wraps all that complexity up in a simple API... `startAt` decides
  where in the array of expressions to start doing the matching. It
  auto-increments, so if a match is found at position 2, then startAt will be
  set to 3.  If the end is reached startAt will return to 0.

  MOST of the time the parser will be setting startAt manually to 0.
*/
class ResumableMultiRegex {
    _language;
    rules;
    multiRegexes;
    count;
    lastIndex;
    regexIndex;
    constructor(language) {
        this._language = language;
        this.rules = [];
        this.multiRegexes = [];
        this.count = 0;
        this.lastIndex = 0;
        this.regexIndex = 0;
    }
    getMatcher = (index) => {
        if (this.multiRegexes[index])
            return this.multiRegexes[index];
        const matcher = new MultiRegex(this._language);
        this.rules.slice(index).forEach(([re, opts]) => matcher.addRule(re, opts));
        matcher.compile();
        this.multiRegexes[index] = matcher;
        return matcher;
    };
    resumingScanAtSamePosition = () => this.regexIndex !== 0;
    considerAll = () => {
        this.regexIndex = 0;
    };
    addRule = (re, opts) => {
        this.rules.push([re, opts]);
        if (opts.type === 'begin') {
            this.count += 1;
        }
    };
    exec = (str) => {
        const match = this.getMatcher(this.regexIndex);
        match.lastIndex = this.lastIndex;
        let result = match.exec(str);
        if (this.resumingScanAtSamePosition()) {
            if (result && result.index === this.lastIndex) ;
            else {
                const match2 = this.getMatcher(0);
                match2.lastIndex = this.lastIndex + 1;
                result = match2.exec(str);
            }
        }
        if (result) {
            this.regexIndex += result.position + 1;
            if (this.regexIndex === this.count) {
                this.considerAll();
            }
        }
        return result;
    };
}

/**
 * Determines if a mode has a dependency on it's parent or not.
 *
 * If a mode does have a parent dependency then often we need to clone it if
 * it's used in multiple places so that each copy points to the correct parent,
 * where-as modes without a parent can often safely be re-used at the bottom of
 * a mode chain.
 *
 * @param {Mode | null} mode
 * @returns {boolean} - is there a dependency on the parent?
 * */
const dependencyOnParent = (mode) => {
    if (!mode)
        return false;
    return mode.endsWithParent || dependencyOnParent(mode.starts);
};
/**
 * Expands a mode or clones it if necessary
 *
 * This is necessary for modes with parental dependenceis (see notes on
 * `dependencyOnParent`) and for nodes that have `variants` - which must then be
 * exploded into their own individual modes at compile time.
 *
 * @param {Mode} mode
 * @returns {Mode | Mode[]}
 * */
const expandOrCloneMode = (mode) => {
    if (mode.variants && !mode.cachedVariants) {
        mode.cachedVariants = mode.variants.map((variant) => inherit(mode, { variants: null }, variant));
    }
    if (mode.cachedVariants) {
        return mode.cachedVariants;
    }
    if (dependencyOnParent(mode)) {
        return inherit(mode, { starts: mode.starts ? inherit(mode.starts) : null });
    }
    if (Object.isFrozen(mode)) {
        return inherit(mode);
    }
    // no special dependency issues, just return ourselves
    return mode;
};
/**
 * Compiles a language definition result.
 *
 * Given the raw result of a language definition (Language), compiles this so
 * that it is ready for highlighting code.
 *
 * @param {Language} language
 * @returns {CompiledLanguage}
 */
const compileLanguage = (language) => {
    /**
     * Given a mode, builds a huge ResumableMultiRegex that can be used to walk
     * the content and find matches.
     *
     * @param {CompiledMode} mode
     * @returns {ResumableMultiRegex}
     */
    const buildModeRegex = (mode) => {
        const mm = new ResumableMultiRegex(language);
        mode.contains.forEach((term) => mm.addRule(term.begin, { rule: term, type: 'begin' }));
        if (mode.terminatorEnd) {
            mm.addRule(mode.terminatorEnd, { type: 'end' });
        }
        if (mode.illegal) {
            mm.addRule(mode.illegal, { type: 'illegal' });
        }
        return mm;
    };
    /**
     * Compiles an individual mode
     *
     * This can raise an error if the mode contains certain detectable known logic
     * issues.
     * @param {Mode} mode
     * @param {CompiledMode | null} parent
     * @returns {CompiledMode | never}
     */
    const compileMode = (mode, parent) => {
        const cMode = mode;
        if (mode.isCompiled)
            return cMode;
        [
            scopeClassName,
            compileMatch,
            multiClass,
            beforeMatchExt,
        ].forEach((ext) => ext(mode, parent));
        language.compilerExtensions.forEach((ext) => ext(mode, parent));
        mode.__beforeBegin = null;
        [
            beginKeywords,
            compileIllegal,
            compileRelevance,
        ].forEach((ext) => ext(mode, parent));
        mode.isCompiled = true;
        let keywordPattern = null;
        if (typeof mode.keywords === 'object' && mode.keywords.$pattern) {
            mode.keywords = { ...mode.keywords };
            keywordPattern = mode.keywords.$pattern;
            delete mode.keywords.$pattern;
        }
        keywordPattern = keywordPattern || /\w+/;
        if (mode.keywords) {
            mode.keywords = compileKeywords(mode.keywords, language.caseInsensitive);
        }
        cMode.keywordPatternRe = langRe(keywordPattern, true, language);
        if (parent) {
            if (!mode.begin) {
                mode.begin = /\B|\b/;
            }
            cMode.beginRe = langRe(cMode.begin, false, language);
            if (!mode.end && !mode.endsWithParent) {
                mode.end = /\B|\b/;
            }
            if (mode.end) {
                cMode.endRe = langRe(cMode.end, false, language);
            }
            cMode.terminatorEnd = source(cMode.end) || '';
            if (mode.endsWithParent && parent.terminatorEnd) {
                cMode.terminatorEnd += (mode.end ? '|' : '') + parent.terminatorEnd;
            }
        }
        if (mode.illegal) {
            cMode.illegalRe = langRe(mode.illegal, false, language);
        }
        if (!mode.contains) {
            mode.contains = [];
        }
        mode.contains = [].concat(...mode.contains.map((c) => expandOrCloneMode(c === 'self' ? mode : c)));
        mode.contains.forEach((c) => { compileMode(c, cMode); });
        if (mode.starts) {
            compileMode(mode.starts, parent);
        }
        cMode.matcher = buildModeRegex(cMode);
        return cMode;
    };
    if (!language.compilerExtensions) {
        language.compilerExtensions = [];
    }
    if (language.contains && language.contains.includes('self')) {
        throw new Error('ERR: contains `self` is not supported at the top-level of a language.  See documentation.');
    }
    language.classNameAliases = inherit(language.classNameAliases || {});
    return compileMode(language);
};

const SPAN_CLOSE = '</span>';
const emitsWrappingTags = (node) => !!node.kind;
const expandScopeName = (name, { prefix }) => {
    if (name.includes('.')) {
        const pieces = name.split('.');
        return [
            `${prefix}${pieces.shift()}`,
            ...(pieces.map((x, i) => `${x}${'_'.repeat(i + 1)}`)),
        ].join(' ');
    }
    return `${prefix}${name}`;
};
class HTMLRenderer {
    buffer;
    classPrefix;
    _options;
    constructor(tree, options) {
        this.buffer = '';
        this.classPrefix = options.classPrefix;
        this._options = options;
        tree.walk(this);
    }
    /**
     * Add texts to the output stream.
     *
     * @param text
     */
    addText = (text) => {
        this.buffer += this._options.escapeHTML ? escapeHTML(text) : text;
    };
    /**
     * Adds a node open to the output stream (if needed).
     *
     * @param {Partial<Node>} node */
    openNode = (node) => {
        if (!emitsWrappingTags(node))
            return undefined;
        let scope = node.kind;
        if (node.subLanguage) {
            scope = `language-${scope}`;
        }
        else {
            scope = expandScopeName(scope, { prefix: this.classPrefix });
        }
        this.span(scope);
        return undefined;
    };
    /**
     * Adds a node close to the output stream (if needed).
     *
     * @param {Partial<Node>} node */
    closeNode = (node) => {
        if (!emitsWrappingTags(node))
            return undefined;
        this.buffer += SPAN_CLOSE;
        return undefined;
    };
    /**
     * returns the accumulated buffer.
     */
    value = () => this.buffer;
    /**
     * Builds a span element
     *
     * @param className
     */
    span = (className) => {
        this.buffer += `<span class="${className}">`;
    };
}

class TokenTree {
    rootNode;
    stack;
    constructor() {
        this.rootNode = { children: [] };
        this.stack = [this.rootNode];
    }
    get top() {
        return this.stack[this.stack.length - 1];
    }
    get root() {
        return this.rootNode;
    }
    add = (node) => {
        this.top.children.push(node);
    };
    openNode = (kind) => {
        const node = { kind, children: [] };
        this.add(node);
        this.stack.push(node);
    };
    closeNode = () => {
        if (this.stack.length > 1)
            return this.stack.pop();
        return undefined;
    };
    closeAllNodes = () => {
        while (this.closeNode())
            ;
    };
    toJSON = () => JSON.stringify(this.rootNode, null, 2);
    walk = (builder) => TokenTree._walk(builder, this.rootNode);
    static _walk(builder, node) {
        if (typeof node === 'string') {
            builder.addText(node);
        }
        else if (node.children) {
            builder.openNode(node);
            node.children.forEach((child) => this._walk(builder, child));
            builder.closeNode(node);
        }
        return builder;
    }
    static _collpase(node) {
        if (typeof node === 'string')
            return undefined;
        if (!node.children)
            return undefined;
        if (node.children.every((el) => typeof el === 'string')) {
            node.children = [node.children.join('')];
        }
        else {
            node.children.forEach((child) => {
                TokenTree._collpase(child);
            });
        }
        return undefined;
    }
}

class TokenTreeEmitter extends TokenTree {
    options;
    constructor(options) {
        super();
        this.options = options;
    }
    addKeyword = (text, kind) => {
        if (text === '')
            return undefined;
        this.openNode(kind);
        this.addText(text);
        this.closeNode();
        return undefined;
    };
    addText = (text) => {
        if (text === '')
            return undefined;
        this.add(text);
        return undefined;
    };
    addSubLanguage = (emitter, name) => {
        const node = emitter.root;
        node.kind = name;
        node.subLanguage = true;
        this.add(node);
    };
    toHTML = () => {
        const renderer = new HTMLRenderer(this, this.options);
        return renderer.value();
    };
    // eslint-disable-next-line class-methods-use-this
    finalize = () => true;
}

class Response {
    data;
    isMatchIgnored;
    constructor(mode) {
        if (mode.data === undefined) {
            mode.data = {};
        }
        this.data = mode.data;
        this.isMatchIgnored = false;
    }
    ignoreMatch = () => {
        this.isMatchIgnored = true;
    };
}

const modFn = (type) => (() => { throw new Error(`${type} is read-only`); });
const deepFreeze = (obj) => {
    if (obj instanceof Map) {
        const _modFn = modFn('Map');
        obj.clear = _modFn;
        obj.delete = _modFn;
        obj.set = _modFn;
    }
    else if (obj instanceof Set) {
        const _modFn = modFn('Set');
        obj.add = _modFn;
        obj.clear = _modFn;
        obj.delete = _modFn;
    }
    Object.freeze(obj);
    Object.getOwnPropertyNames(obj).forEach((key) => {
        const prop = obj[key];
        if (typeof prop === 'object' && !Object.isFrozen(obj)) {
            deepFreeze(prop);
        }
    });
    return obj;
};

const IDENT_RE$1 = '[a-zA-Z]\\w*';
const UNDERSCORE_IDENT_RE = '[a-zA-Z_]\\w*';
const NUMBER_RE = '\\b\\d+(\\.\\d+)?';
const C_NUMBER_RE = '(-?)(\\b0[xX][a-fA-F0-9]+|(\\b\\d+(\\.\\d*)?|\\.\\d+)([eE][-+]?\\d+)?)';
const RE_STARTERS_RE = '!|!=|!==|%|%=|&|&&|&=|\\*|\\*=|\\+|\\+=|,|-|-=|/=|/|:|;|<<|<<=|<=|<|===|==|=|>>>=|>>=|>=|>>>|>>|>|\\?|\\[|\\{|\\(|\\^|\\^=|\\||\\|=|\\|\\||~';
const SHEBANG = (opts = {}) => {
    const beginShebang = /^#![ ]*\//;
    if (opts.binary) {
        opts.begin = concat(beginShebang, /.*\b/, opts.binary, /\b.*/);
    }
    return inherit({
        scope: 'meta',
        begin: beginShebang,
        end: /$/,
        relevance: 0,
        'on:begin': (match, response) => match.index !== 0 && response.ignoreMatch(),
    }, opts);
};
// Common modes
const _BACKSLASH_ESCAPE = { begin: '\\\\[\\s\\S]', relevance: 0 };
const BACKSLASH_ESCAPE = deepFreeze(_BACKSLASH_ESCAPE);
const _APOS_STRING_MODE = {
    scope: 'string',
    begin: '\'',
    end: '\'',
    illegal: '\\n',
    contains: [BACKSLASH_ESCAPE],
};
const APOS_STRING_MODE = deepFreeze(_APOS_STRING_MODE);
const _QUOTE_STRING_MODE = {
    scope: 'string',
    begin: '"',
    end: '"',
    illegal: '\\n',
    contains: [BACKSLASH_ESCAPE],
};
const QUOTE_STRING_MODE = deepFreeze(_QUOTE_STRING_MODE);
/**
 * Creates a comment mode.
 *
 * @param {string | RegExp} begin
 * @param {string | RegExp} end
 * @param {string | RegExp} modeOptions
 */
const COMMENT = (begin, end, modeOptions = {}) => {
    const mode = inherit({
        scope: 'comment',
        begin,
        end,
        contains: [],
    }, modeOptions);
    mode.contains?.push({
        scope: 'doctag',
        // hack to avoid the space from being included. the space is necessary to
        // match here to prevent the plain text rule below from gobbling up doctags
        begin: '[ ]*(?=(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):)',
        end: /(TODO|FIXME|NOTE|BUG|OPTIMIZE|HACK|XXX):/,
        excludeBegin: true,
        relevance: 0,
    });
    const ENGLISH_WORD = either(
    // list of common 1 and 2 letter words in English
    'I', 'a', 'is', 'so', 'us', 'to', 'at', 'if', 'in', 'it', 'on', 
    // note: this is not an exhaustive list of contractions, just popular ones
    /[A-Za-z]+['](d|ve|re|ll|t|s|n)/, // contractions - can't we'd they're let's, etc
    /[A-Za-z]+[-][a-z]+/, // `no-way`, etc.
    /[A-Za-z][a-z]{2,}/);
    // looking like plain text, more likely to be a comment
    mode.contains?.push({
        begin: concat(/[ ]+/, // necessary to prevent us gobbling up doctags like /* @author Bob Mcgill */
        '(', ENGLISH_WORD, /[.]?[:]?([.][ ]|[ ])/, '){3}'),
    });
    return mode;
};
const C_LINE_COMMENT_MODE = COMMENT('//', '$');
const C_BLOCK_COMMENT_MODE = COMMENT('/\\*', '\\*/');
const HASH_COMMENT_MODE = COMMENT('#', '$');
const C_NUMBER_MODE = {
    scope: 'number',
    begin: C_NUMBER_RE,
    relevance: 0,
};
const _REGEXP_MODE = {
    begin: /(?=\/[^/\n]*\/)/,
    contains: [{
            scope: 'regexp',
            begin: /\//,
            end: /\/[gimuy]*/,
            illegal: /\n/,
            contains: [
                BACKSLASH_ESCAPE,
                {
                    begin: /\[/,
                    end: /\]/,
                    relevance: 0,
                    contains: [BACKSLASH_ESCAPE],
                },
            ],
        }],
};
const REGEXP_MODE = deepFreeze(_REGEXP_MODE);
const TITLE_MODE = {
    scope: 'title',
    begin: IDENT_RE$1,
    relevance: 0,
};

const CSS_MODES = {
    IMPORTANT: { scope: 'meta', begin: '!important' },
    HEX_COLOR: { scope: 'number', begin: /#(([0-9a-fA-F]{3,4})|(([0-9a-fA-F]{2}){3,4}))\b/ },
    FUNCTION_DISPATCH: { className: 'built_in', begin: /[\w-]+(?=\()/ },
    ATTRIBUTE_SELECTOR: {
        scope: 'selector-attr',
        begin: /\[/,
        end: /\]/,
        illegal: '$',
        contains: [APOS_STRING_MODE, QUOTE_STRING_MODE],
    },
    NUMBER: {
        scope: 'number',
        begin: [
            NUMBER_RE,
            '(',
            '%|em|ex|ch|rem',
            '|vw|vh|vmin|vmax',
            '|cm|mm|in|pt|pc|px',
            '|deg|grad|rad|turn',
            '|s|ms',
            '|Hz|kHz',
            '|dpi|dpcm|dppx',
            ')?',
        ].join(''),
        relevance: 0,
    },
    VARIABLE: { className: 'attr', begin: /--[A-Za-z][A-Za-z0-9_-]*/ },
};
const TAGS = [
    'a',
    'abbr',
    'address',
    'article',
    'aside',
    'audio',
    'b',
    'blockquote',
    'body',
    'button',
    'canvas',
    'caption',
    'cite',
    'code',
    'dd',
    'del',
    'details',
    'dfn',
    'div',
    'dl',
    'dt',
    'em',
    'fieldset',
    'figcaption',
    'figure',
    'footer',
    'form',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6',
    'header',
    'hgroup',
    'html',
    'i',
    'iframe',
    'img',
    'input',
    'ins',
    'kbd',
    'label',
    'legend',
    'li',
    'main',
    'mark',
    'menu',
    'nav',
    'object',
    'ol',
    'p',
    'q',
    'quote',
    'samp',
    'section',
    'span',
    'strong',
    'summary',
    'sup',
    'table',
    'tbody',
    'td',
    'textarea',
    'tfoot',
    'th',
    'thead',
    'time',
    'tr',
    'ul',
    'var',
    'video',
];
const MEDIA_FEATURES = [
    'any-hover',
    'any-pointer',
    'aspect-ratio',
    'color',
    'color-gamut',
    'color-index',
    'device-aspect-ratio',
    'device-height',
    'device-width',
    'display-mode',
    'forced-colors',
    'grid',
    'height',
    'hover',
    'inverted-colors',
    'monochrome',
    'orientation',
    'overflow-block',
    'overflow-inline',
    'pointer',
    'prefers-color-scheme',
    'prefers-contrast',
    'prefers-reduced-motion',
    'prefers-reduced-transparency',
    'resolution',
    'scan',
    'scripting',
    'update',
    'width',
    // TODO: find a better solution?
    'min-width',
    'max-width',
    'min-height',
    'max-height',
];
// https://developer.mozilla.org/en-US/docs/Web/CSS/Pseudo-classes
const PSEUDO_CLASSES = [
    'active',
    'any-link',
    'blank',
    'checked',
    'current',
    'default',
    'defined',
    'dir',
    'disabled',
    'drop',
    'empty',
    'enabled',
    'first',
    'first-child',
    'first-of-type',
    'fullscreen',
    'future',
    'focus',
    'focus-visible',
    'focus-within',
    'has',
    'host',
    'host-context',
    'hover',
    'indeterminate',
    'in-range',
    'invalid',
    'is',
    'lang',
    'last-child',
    'last-of-type',
    'left',
    'link',
    'local-link',
    'not',
    'nth-child',
    'nth-col',
    'nth-last-child',
    'nth-last-col',
    'nth-last-of-type',
    'nth-of-type',
    'only-child',
    'only-of-type',
    'optional',
    'out-of-range',
    'past',
    'placeholder-shown',
    'read-only',
    'read-write',
    'required',
    'right',
    'root',
    'scope',
    'target',
    'target-within',
    'user-invalid',
    'valid',
    'visited',
    'where', // where()
];
// https://developer.mozilla.org/en-US/docs/Web/CSS/Pseudo-elements
const PSEUDO_ELEMENTS = [
    'after',
    'backdrop',
    'before',
    'cue',
    'cue-region',
    'first-letter',
    'first-line',
    'grammar-error',
    'marker',
    'part',
    'placeholder',
    'selection',
    'slotted',
    'spelling-error',
];
const ATTRIBUTES = [
    'align-content',
    'align-items',
    'align-self',
    'all',
    'animation',
    'animation-delay',
    'animation-direction',
    'animation-duration',
    'animation-fill-mode',
    'animation-iteration-count',
    'animation-name',
    'animation-play-state',
    'animation-timing-function',
    'backface-visibility',
    'background',
    'background-attachment',
    'background-blend-mode',
    'background-clip',
    'background-color',
    'background-image',
    'background-origin',
    'background-position',
    'background-repeat',
    'background-size',
    'block-size',
    'border',
    'border-block',
    'border-block-color',
    'border-block-end',
    'border-block-end-color',
    'border-block-end-style',
    'border-block-end-width',
    'border-block-start',
    'border-block-start-color',
    'border-block-start-style',
    'border-block-start-width',
    'border-block-style',
    'border-block-width',
    'border-bottom',
    'border-bottom-color',
    'border-bottom-left-radius',
    'border-bottom-right-radius',
    'border-bottom-style',
    'border-bottom-width',
    'border-collapse',
    'border-color',
    'border-image',
    'border-image-outset',
    'border-image-repeat',
    'border-image-slice',
    'border-image-source',
    'border-image-width',
    'border-inline',
    'border-inline-color',
    'border-inline-end',
    'border-inline-end-color',
    'border-inline-end-style',
    'border-inline-end-width',
    'border-inline-start',
    'border-inline-start-color',
    'border-inline-start-style',
    'border-inline-start-width',
    'border-inline-style',
    'border-inline-width',
    'border-left',
    'border-left-color',
    'border-left-style',
    'border-left-width',
    'border-radius',
    'border-right',
    'border-right-color',
    'border-right-style',
    'border-right-width',
    'border-spacing',
    'border-style',
    'border-top',
    'border-top-color',
    'border-top-left-radius',
    'border-top-right-radius',
    'border-top-style',
    'border-top-width',
    'border-width',
    'bottom',
    'box-decoration-break',
    'box-shadow',
    'box-sizing',
    'break-after',
    'break-before',
    'break-inside',
    'caption-side',
    'caret-color',
    'clear',
    'clip',
    'clip-path',
    'clip-rule',
    'color',
    'column-count',
    'column-fill',
    'column-gap',
    'column-rule',
    'column-rule-color',
    'column-rule-style',
    'column-rule-width',
    'column-span',
    'column-width',
    'columns',
    'contain',
    'content',
    'content-visibility',
    'counter-increment',
    'counter-reset',
    'cue',
    'cue-after',
    'cue-before',
    'cursor',
    'direction',
    'display',
    'empty-cells',
    'filter',
    'flex',
    'flex-basis',
    'flex-direction',
    'flex-flow',
    'flex-grow',
    'flex-shrink',
    'flex-wrap',
    'float',
    'flow',
    'font',
    'font-display',
    'font-family',
    'font-feature-settings',
    'font-kerning',
    'font-language-override',
    'font-size',
    'font-size-adjust',
    'font-smoothing',
    'font-stretch',
    'font-style',
    'font-synthesis',
    'font-variant',
    'font-variant-caps',
    'font-variant-east-asian',
    'font-variant-ligatures',
    'font-variant-numeric',
    'font-variant-position',
    'font-variation-settings',
    'font-weight',
    'gap',
    'glyph-orientation-vertical',
    'grid',
    'grid-area',
    'grid-auto-columns',
    'grid-auto-flow',
    'grid-auto-rows',
    'grid-column',
    'grid-column-end',
    'grid-column-start',
    'grid-gap',
    'grid-row',
    'grid-row-end',
    'grid-row-start',
    'grid-template',
    'grid-template-areas',
    'grid-template-columns',
    'grid-template-rows',
    'hanging-punctuation',
    'height',
    'hyphens',
    'icon',
    'image-orientation',
    'image-rendering',
    'image-resolution',
    'ime-mode',
    'inline-size',
    'isolation',
    'justify-content',
    'left',
    'letter-spacing',
    'line-break',
    'line-height',
    'list-style',
    'list-style-image',
    'list-style-position',
    'list-style-type',
    'margin',
    'margin-block',
    'margin-block-end',
    'margin-block-start',
    'margin-bottom',
    'margin-inline',
    'margin-inline-end',
    'margin-inline-start',
    'margin-left',
    'margin-right',
    'margin-top',
    'marks',
    'mask',
    'mask-border',
    'mask-border-mode',
    'mask-border-outset',
    'mask-border-repeat',
    'mask-border-slice',
    'mask-border-source',
    'mask-border-width',
    'mask-clip',
    'mask-composite',
    'mask-image',
    'mask-mode',
    'mask-origin',
    'mask-position',
    'mask-repeat',
    'mask-size',
    'mask-type',
    'max-block-size',
    'max-height',
    'max-inline-size',
    'max-width',
    'min-block-size',
    'min-height',
    'min-inline-size',
    'min-width',
    'mix-blend-mode',
    'nav-down',
    'nav-index',
    'nav-left',
    'nav-right',
    'nav-up',
    'none',
    'normal',
    'object-fit',
    'object-position',
    'opacity',
    'order',
    'orphans',
    'outline',
    'outline-color',
    'outline-offset',
    'outline-style',
    'outline-width',
    'overflow',
    'overflow-wrap',
    'overflow-x',
    'overflow-y',
    'padding',
    'padding-block',
    'padding-block-end',
    'padding-block-start',
    'padding-bottom',
    'padding-inline',
    'padding-inline-end',
    'padding-inline-start',
    'padding-left',
    'padding-right',
    'padding-top',
    'page-break-after',
    'page-break-before',
    'page-break-inside',
    'pause',
    'pause-after',
    'pause-before',
    'perspective',
    'perspective-origin',
    'pointer-events',
    'position',
    'quotes',
    'resize',
    'rest',
    'rest-after',
    'rest-before',
    'right',
    'row-gap',
    'scroll-margin',
    'scroll-margin-block',
    'scroll-margin-block-end',
    'scroll-margin-block-start',
    'scroll-margin-bottom',
    'scroll-margin-inline',
    'scroll-margin-inline-end',
    'scroll-margin-inline-start',
    'scroll-margin-left',
    'scroll-margin-right',
    'scroll-margin-top',
    'scroll-padding',
    'scroll-padding-block',
    'scroll-padding-block-end',
    'scroll-padding-block-start',
    'scroll-padding-bottom',
    'scroll-padding-inline',
    'scroll-padding-inline-end',
    'scroll-padding-inline-start',
    'scroll-padding-left',
    'scroll-padding-right',
    'scroll-padding-top',
    'scroll-snap-align',
    'scroll-snap-stop',
    'scroll-snap-type',
    'scrollbar-color',
    'scrollbar-gutter',
    'scrollbar-width',
    'shape-image-threshold',
    'shape-margin',
    'shape-outside',
    'speak',
    'speak-as',
    'src',
    'tab-size',
    'table-layout',
    'text-align',
    'text-align-all',
    'text-align-last',
    'text-combine-upright',
    'text-decoration',
    'text-decoration-color',
    'text-decoration-line',
    'text-decoration-style',
    'text-emphasis',
    'text-emphasis-color',
    'text-emphasis-position',
    'text-emphasis-style',
    'text-indent',
    'text-justify',
    'text-orientation',
    'text-overflow',
    'text-rendering',
    'text-shadow',
    'text-transform',
    'text-underline-position',
    'top',
    'transform',
    'transform-box',
    'transform-origin',
    'transform-style',
    'transition',
    'transition-delay',
    'transition-duration',
    'transition-property',
    'transition-timing-function',
    'unicode-bidi',
    'vertical-align',
    'visibility',
    'voice-balance',
    'voice-duration',
    'voice-family',
    'voice-pitch',
    'voice-range',
    'voice-rate',
    'voice-stress',
    'voice-volume',
    'white-space',
    'widows',
    'width',
    'will-change',
    'word-break',
    'word-spacing',
    'word-wrap',
    'writing-mode',
    'z-index',
    // reverse makes sure longer attributes `font-weight` are matched fully
    // instead of getting false positives on say `font`
].reverse();

const cssLanguage = () => {
    const VENDOR_PREFIX = { begin: /-(webkit|moz|ms|o)-(?=[a-z])/ };
    const AT_MODIFIERS = 'and or not only';
    const AT_PROPERTY_RE = /@-?\w[\w]*(-\w+)*/; // @-webkit-keyframes
    const IDENT_RE = '[a-zA-Z-][a-zA-Z0-9_-]*';
    const STRINGS = [APOS_STRING_MODE, QUOTE_STRING_MODE];
    return {
        name: 'css',
        caseInsensitive: true,
        illegal: /[=|'$]/,
        keywords: { keyframePosition: 'from to' },
        classNameAliases: { keyframePosition: 'selector-tag' },
        contains: [
            C_BLOCK_COMMENT_MODE,
            VENDOR_PREFIX,
            CSS_MODES.NUMBER,
            { className: 'selector-id', begin: /#[A-Za-z0-9_-]+/, relevance: 0 },
            { className: 'selector-class', begin: `\\.${IDENT_RE}`, relevance: 0 },
            CSS_MODES.ATTRIBUTE_SELECTOR,
            {
                className: 'selector-pseudo',
                variants: [
                    { begin: `:(${PSEUDO_CLASSES.join('|')})` },
                    { begin: `:(:)?(${PSEUDO_ELEMENTS.join('|')})` },
                ],
            },
            CSS_MODES.VARIABLE,
            { className: 'attribute', begin: `\\b(${ATTRIBUTES.join('|')})\\b` },
            {
                begin: /:/,
                end: /[;}{]/,
                contains: [
                    C_BLOCK_COMMENT_MODE,
                    CSS_MODES.HEX_COLOR,
                    CSS_MODES.IMPORTANT,
                    CSS_MODES.NUMBER,
                    ...STRINGS,
                    {
                        begin: /(url|data-uri)\(/,
                        end: /\)/,
                        relevance: 0,
                        keywords: { built_in: 'url data-uri' },
                        contains: [{
                                className: 'string',
                                begin: /[^)]/,
                                endsWithParent: true,
                                excludeEnd: true,
                            }],
                    },
                    CSS_MODES.FUNCTION_DISPATCH,
                ],
            },
            {
                begin: lookahead(/@/),
                end: '[{;]',
                relevance: 0,
                illegal: /:/,
                contains: [
                    { className: 'keyword', begin: AT_PROPERTY_RE },
                    {
                        begin: /\s/,
                        endsWithParent: true,
                        excludeEnd: true,
                        relevance: 0,
                        keywords: {
                            $pattern: /[a-z-]+/,
                            keyword: AT_MODIFIERS,
                            attribute: MEDIA_FEATURES.join(' '),
                        },
                        contains: [
                            { begin: /[a-z-]+(?=:)/, className: 'attribute' },
                            ...STRINGS,
                            CSS_MODES.NUMBER,
                        ],
                    },
                ],
            },
            { className: 'selector-tag', begin: `\\b(${TAGS.join('|')})\\b` },
        ],
    };
};

const IDENT_RE = '[A-Za-z$_][0-9A-Za-z$_]*';
const KEYWORDS = [
    'as',
    'in',
    'of',
    'if',
    'for',
    'while',
    'finally',
    'var',
    'new',
    'function',
    'do',
    'return',
    'void',
    'else',
    'break',
    'catch',
    'instanceof',
    'with',
    'throw',
    'case',
    'default',
    'try',
    'switch',
    'continue',
    'typeof',
    'delete',
    'let',
    'yield',
    'const',
    'class',
    // JS handles these with a special rule
    // 'get',
    // 'set',
    'debugger',
    'async',
    'await',
    'static',
    'import',
    'from',
    'export',
    'extends',
];
const LITERALS = [
    'true',
    'false',
    'null',
    'undefined',
    'NaN',
    'Infinity',
];
const TYPES = [
    // Fundamental objects
    'Object',
    'Function',
    'Boolean',
    'Symbol',
    // numbers and dates
    'Math',
    'Date',
    'Number',
    'BigInt',
    // text
    'String',
    'RegExp',
    // Indexed collections
    'Array',
    'Float32Array',
    'Float64Array',
    'Int8Array',
    'Uint8Array',
    'Uint8ClampedArray',
    'Int16Array',
    'Int32Array',
    'Uint16Array',
    'Uint32Array',
    'BigInt64Array',
    'BigUint64Array',
    // Keyed collections
    'Set',
    'Map',
    'WeakSet',
    'WeakMap',
    // Structured data
    'ArrayBuffer',
    'SharedArrayBuffer',
    'Atomics',
    'DataView',
    'JSON',
    // Control abstraction objects
    'Promise',
    'Generator',
    'GeneratorFunction',
    'AsyncFunction',
    // Reflection
    'Reflect',
    'Proxy',
    // Internationalization
    'Intl',
    // WebAssembly
    'WebAssembly',
];
const ERROR_TYPES = [
    'Error',
    'EvalError',
    'InternalError',
    'RangeError',
    'ReferenceError',
    'SyntaxError',
    'TypeError',
    'URIError',
];
const BUILT_IN_GLOBALS = [
    'setInterval',
    'setTimeout',
    'clearInterval',
    'clearTimeout',
    'require',
    'exports',
    'eval',
    'isFinite',
    'isNaN',
    'parseFloat',
    'parseInt',
    'decodeURI',
    'decodeURIComponent',
    'encodeURI',
    'encodeURIComponent',
    'escape',
    'unescape',
];
const BUILT_IN_VARIABLES = [
    'arguments',
    'this',
    'super',
    'console',
    'window',
    'document',
    'localStorage',
    'module',
    'global', // Node.js
];
const _ARR = [];
const BUILT_INS = _ARR.concat(BUILT_IN_GLOBALS, TYPES, ERROR_TYPES);

const javascriptLanguage = () => {
    const hasClosingTag = (match, { after }) => (match.input?.indexOf(`</${match[0].slice(1)}`, after) !== -1);
    const FRAGMENT = {
        begin: '<>',
        end: '</>',
    };
    // to avoid some special cases inside isTrulyOpeningTag
    const XML_SELF_CLOSING = /<[A-Za-z0-9\\._:-]+\s*\/>/;
    const XML_TAG = {
        begin: /<[A-Za-z0-9\\._:-]+/,
        end: /\/[A-Za-z0-9\\._:-]+>|\/>/,
        isTrulyOpeningTag: (match, response) => {
            const afterMatchIndex = match[0].length + (match.index || 0);
            const nextChar = match.input && match.input[afterMatchIndex];
            // HTML should not include another raw `<` inside a tag
            // nested type?
            // `<Array<Array<number>>`, etc.
            // the , gives away that this is not HTML
            // `<T, A extends keyof T, V>`
            if (nextChar === '<' || nextChar === ',')
                return response.ignoreMatch();
            // `<something>`
            // Quite possibly a tag, lets look for a matching closing tag...
            if (nextChar === '>') {
                if (!hasClosingTag(match, { after: afterMatchIndex })) {
                    response.ignoreMatch();
                }
                // `<blah />` (self-closing)
                // handled by simpleSelfClosing rule
                // `<From extends string>`
                // technically this could be HTML, but it smells like a type
                const afterMatch = match.input?.substring(afterMatchIndex);
                const m = afterMatch?.match(/^\s+extends\s+/);
                // NOTE: This is ugh, but added specifically for https://github.com/highlightjs/highlight.js/issues/3276
                if (m && m.index === 0) {
                    return response.ignoreMatch();
                }
            }
            return undefined;
        },
    };
    const KEYWORDS$1 = {
        $pattern: IDENT_RE,
        keyword: KEYWORDS,
        literal: LITERALS,
        built_in: BUILT_INS,
        'variable.language': BUILT_IN_VARIABLES,
    };
    const decimalDigits = '[0-9](_?[0-9])*';
    const frac = `\\.(${decimalDigits})`;
    const decimalInteger = '0|[1-9](_?[0-9])*|0[0-7]*[89][0-9]*';
    const NUMBER = {
        className: 'number',
        variants: [
            {
                begin: [
                    `(\\b(${decimalInteger})((${frac})|\\.)?|(${frac}))`,
                    `[eE][+-]?(${decimalDigits})\\b`,
                ].join(''),
            },
            { begin: `\\b(${decimalInteger})\\b((${frac})\\b|\\.)?|(${frac})\\b` },
            { begin: '\\b(0|[1-9](_?[0-9])*)n\\b' },
            { begin: '\\b0[xX][0-9a-fA-F](_?[0-9a-fA-F])*n?\\b' },
            { begin: '\\b0[bB][0-1](_?[0-1])*n?\\b' },
            { begin: '\\b0[oO][0-7](_?[0-7])*n?\\b' },
            { begin: '\\b0[0-7]+n?\\b' },
        ],
        relevance: 0,
    };
    const SUBST = {
        className: 'subst',
        begin: '\\$\\{',
        end: '\\}',
        keywords: KEYWORDS$1,
        contains: [], // defined later
    };
    const HTML_TEMPLATE = {
        begin: 'html`',
        end: '',
        starts: {
            end: '`',
            returnEnd: false,
            contains: [BACKSLASH_ESCAPE, SUBST],
            subLanguage: 'xml',
        },
    };
    const CSS_TEMPLATE = {
        begin: 'css`',
        end: '',
        starts: {
            end: '`',
            returnEnd: false,
            contains: [BACKSLASH_ESCAPE, SUBST],
            subLanguage: 'css',
        },
    };
    const TEMPLATE_STRING = {
        className: 'string',
        begin: '`',
        end: '`',
        contains: [BACKSLASH_ESCAPE, SUBST],
    };
    const JSDOC_COMMENT = COMMENT(/\/\*\*(?!\/)/, '\\*/', {
        relevance: 0,
        contains: [
            {
                begin: '(?=@[A-Za-z]+)',
                relevance: 0,
                contains: [
                    { className: 'doctag', begin: '@[A-Za-z]+' },
                    {
                        className: 'type',
                        begin: '\\{',
                        end: '\\}',
                        excludeEnd: true,
                        excludeBegin: true,
                        relevance: 0,
                    },
                    {
                        className: 'variable',
                        begin: `${IDENT_RE}(?=\\s*(-)|$)`,
                        endsParent: true,
                        relevance: 0,
                    },
                    { begin: /(?=[^\n])\s/, relevance: 0 },
                ],
            },
        ],
    });
    const JS_COMMENT = {
        className: 'comment',
        variants: [
            JSDOC_COMMENT,
            C_BLOCK_COMMENT_MODE,
            C_LINE_COMMENT_MODE,
        ],
    };
    const SUBST_INTERNALS = [
        APOS_STRING_MODE,
        QUOTE_STRING_MODE,
        HTML_TEMPLATE,
        CSS_TEMPLATE,
        TEMPLATE_STRING,
        NUMBER,
    ];
    SUBST.contains = SUBST_INTERNALS.concat({
        begin: /\{/,
        end: /\}/,
        keywords: KEYWORDS$1,
        contains: ['self', ...SUBST_INTERNALS],
    });
    const SUBST_AND_COMMENTS = [{}].filter((i) => Object.keys(i).length)
        .concat(JS_COMMENT, SUBST.contains);
    const PARAMS_CONTAINS = SUBST_AND_COMMENTS.concat([{
            begin: /\(/,
            end: /\)/,
            keywords: KEYWORDS$1,
            contains: ['self', ...SUBST_AND_COMMENTS],
        }]);
    const PARAMS = {
        className: 'params',
        begin: /\(/,
        end: /\)/,
        excludeBegin: true,
        excludeEnd: true,
        keywords: KEYWORDS$1,
        contains: PARAMS_CONTAINS,
    };
    // ES6 classes
    const CLASS_OR_EXTENDS = {
        variants: [
            // class Car extends vehicle
            {
                match: [
                    /class/,
                    /\s+/,
                    IDENT_RE,
                    /\s+/,
                    /extends/,
                    /\s+/,
                    concat(IDENT_RE, '(', concat(/\./, IDENT_RE), ')*'),
                ],
                scope: {
                    1: 'keyword',
                    3: 'title.class',
                    5: 'keyword',
                    7: 'title.class.inherited',
                },
            },
            // class Car
            {
                match: [
                    /class/,
                    /\s+/,
                    IDENT_RE,
                ],
                scope: { 1: 'keyword', 3: 'title.class' },
            },
        ],
    };
    const CLASS_REFERENCE = {
        relevance: 0,
        match: either(/\bJSON/, /\b[A-Z][a-z]+([A-Z][a-z]*|\d)*/, /\b[A-Z]{2,}([A-Z][a-z]+|\d)+([A-Z][a-z]*)*/, /\b[A-Z]{2,}[a-z]+([A-Z][a-z]+|\d)*([A-Z][a-z]*)*/),
        className: 'title.class',
        keywords: {
            _: [
                ...TYPES,
                ...ERROR_TYPES,
            ],
        },
    };
    const USE_STRICT = {
        label: 'use_strict',
        className: 'meta',
        relevance: 10,
        begin: /^\s*['"]use (strict|asm)['"]/,
    };
    const FUNCTION_DEFINITION = {
        variants: [
            {
                match: [
                    /function/,
                    /\s+/,
                    IDENT_RE,
                    /(?=\s*\()/,
                ],
            },
            // anonymous function
            {
                match: [
                    /function/,
                    /\s*(?=\()/,
                ],
            },
        ],
        scope: {
            1: 'keyword',
            3: 'title.function',
        },
        label: 'func.def',
        contains: [PARAMS],
        illegal: /%/,
    };
    const UPPER_CASE_CONSTANT = {
        relevance: 0,
        match: /\b[A-Z][A-Z_0-9]+\b/,
        className: 'variable.constant',
    };
    const FUNCTION_CALL = {
        match: concat(/\b/, concat('(?!', [...BUILT_IN_GLOBALS, 'super'].join('|'), ')'), IDENT_RE, lookahead(/\(/)),
        className: 'title.function',
        relevance: 0,
    };
    const PROPERTY_ACCESS = {
        begin: concat(/\./, lookahead((concat(IDENT_RE, /(?![0-9A-Za-z$_(])/)))),
        end: IDENT_RE,
        excludeBegin: true,
        keywords: 'prototype',
        className: 'property',
        relevance: 0,
    };
    const GETTER_OR_SETTER = {
        match: [
            /get|set/,
            /\s+/,
            IDENT_RE,
            /(?=\()/,
        ],
        scope: {
            1: 'keyword',
            3: 'title.function',
        },
        contains: [{ begin: /\(\)/ }, PARAMS],
    };
    const FUNC_LEAD_IN_RE = [
        '(\\(',
        '[^()]*(\\(',
        '[^()]*(\\(',
        '[^()]*',
        '\\)[^()]*)*',
        '\\)[^()]*)*',
        `\\)|${UNDERSCORE_IDENT_RE})\\s*=>`,
    ].join('');
    const FUNCTION_VARIABLE = {
        match: [
            /const|var|let/, /\s+/,
            IDENT_RE,
            /\s*/,
            /=\s*/,
            /(async\s*)?/,
            lookahead(FUNC_LEAD_IN_RE),
        ],
        keywords: 'async',
        scope: {
            1: 'keyword',
            3: 'title.function',
        },
        contains: [PARAMS],
    };
    return {
        name: 'Javascript',
        aliases: ['js', 'jsx', 'mjs', 'cjs'],
        keywords: KEYWORDS$1,
        // this will be extended by TypeScript
        exports: { PARAMS_CONTAINS, CLASS_REFERENCE },
        illegal: /#(?![$_A-z])/,
        contains: [
            SHEBANG({
                label: 'shebang',
                binary: 'node',
                relevance: 5,
            }),
            USE_STRICT,
            APOS_STRING_MODE,
            QUOTE_STRING_MODE,
            HTML_TEMPLATE,
            CSS_TEMPLATE,
            TEMPLATE_STRING,
            JS_COMMENT,
            NUMBER,
            CLASS_REFERENCE,
            {
                className: 'attr',
                begin: IDENT_RE + lookahead(':'),
                relevance: 0,
            },
            FUNCTION_VARIABLE,
            {
                begin: `(${RE_STARTERS_RE}|\\b(case|return|throw)\\b)\\s*`,
                keywords: 'return throw case',
                relevance: 0,
                contains: [
                    JS_COMMENT,
                    REGEXP_MODE,
                    {
                        className: 'function',
                        // we have to count the parens to make sure we actually have the
                        // correct bounding ( ) before the =>.  There could be any number of
                        // sub-expressions inside also surrounded by parens.
                        begin: FUNC_LEAD_IN_RE,
                        returnBegin: true,
                        end: '\\s*=>',
                        contains: [{
                                className: 'params',
                                variants: [
                                    { begin: UNDERSCORE_IDENT_RE, relevance: 0 },
                                    { begin: /\(\s*\)/, skip: true },
                                    {
                                        begin: /\(/,
                                        end: /\)/,
                                        excludeBegin: true,
                                        excludeEnd: true,
                                        keywords: KEYWORDS$1,
                                        contains: PARAMS_CONTAINS,
                                    },
                                ],
                            }],
                    },
                    // could be a comma delimited list of params to a function call
                    { begin: /,/, relevance: 0 },
                    { match: /\s+/, relevance: 0 },
                    {
                        variants: [
                            FRAGMENT,
                            { match: XML_SELF_CLOSING },
                            {
                                begin: XML_TAG.begin,
                                'on:begin': XML_TAG.isTrulyOpeningTag,
                                end: XML_TAG.end,
                            },
                        ],
                        subLanguage: 'xml',
                        contains: [{
                                begin: XML_TAG.begin,
                                end: XML_TAG.end,
                                skip: true,
                                contains: ['self'],
                            }],
                    },
                ],
            },
            FUNCTION_DEFINITION,
            // prevent this from getting swallowed up by function
            // since they appear "function like"
            { beginKeywords: 'while if switch catch for' },
            {
                // we have to count the parens to make sure we actually have the correct
                // bounding ( ).  There could be any number of sub-expressions inside
                // also surrounded by parens.
                begin: [
                    '\\b(?!function)',
                    UNDERSCORE_IDENT_RE,
                    '\\(',
                    '[^()]*(\\(',
                    '[^()]*(\\(',
                    '[^()]*',
                    '\\)[^()]*)*',
                    '\\)[^()]*)*',
                    '\\)\\s*\\{', // end parens
                ].join(''),
                returnBegin: true,
                label: 'func.def',
                contains: [
                    PARAMS,
                    inherit(TITLE_MODE, { being: IDENT_RE, className: 'title.function' }),
                ],
            },
            // catch ... so it won't trigger the property rule below
            { match: /\.\.\./, relevance: 0 },
            PROPERTY_ACCESS,
            // hack: prevents detection of keywords in some circumstances
            // .keyword()
            // $keyword = x
            { match: `\\$${IDENT_RE}`, relevance: 0 },
            { match: [/\bconstructor(?=\s*\()/], scope: { 1: 'title.function' }, contains: [PARAMS] },
            FUNCTION_CALL,
            UPPER_CASE_CONSTANT,
            CLASS_OR_EXTENDS,
            GETTER_OR_SETTER,
            // relevance booster for a pattern common to JS libs: `$(something)` and `$.something`
            { match: /\$[(.]/ },
        ],
    };
};

const jsonLanguage = () => {
    const ATTRIBUTE = {
        className: 'attr',
        begin: /"(\\.|[^\\"\r\n])*"(?=\s*:)/,
        relevance: 1.01,
    };
    const PUNCTUATION = {
        match: /[{}[\],:]/,
        className: 'punctuation',
        relevance: 0,
    };
    const LITERALS = { beginKeywords: 'true false null' };
    return {
        name: 'JSON',
        contains: [
            ATTRIBUTE,
            PUNCTUATION,
            QUOTE_STRING_MODE,
            LITERALS,
            C_NUMBER_MODE,
            C_LINE_COMMENT_MODE,
            C_BLOCK_COMMENT_MODE,
        ],
        illegal: '\\S',
    };
};

const perlLanguage = () => {
    const KEYWORDS = [
        'abs',
        'accept',
        'alarm',
        'and',
        'atan2',
        'bind',
        'binmode',
        'bless',
        'break',
        'caller',
        'chdir',
        'chmod',
        'chomp',
        'chop',
        'chown',
        'chr',
        'chroot',
        'close',
        'closedir',
        'connect',
        'continue',
        'cos',
        'crypt',
        'dbmclose',
        'dbmopen',
        'defined',
        'delete',
        'die',
        'do',
        'dump',
        'each',
        'else',
        'elsif',
        'endgrent',
        'endhostent',
        'endnetent',
        'endprotoent',
        'endpwent',
        'endservent',
        'eof',
        'eval',
        'exec',
        'exists',
        'exit',
        'exp',
        'fcntl',
        'fileno',
        'flock',
        'for',
        'foreach',
        'fork',
        'format',
        'formline',
        'getc',
        'getgrent',
        'getgrgid',
        'getgrnam',
        'gethostbyaddr',
        'gethostbyname',
        'gethostent',
        'getlogin',
        'getnetbyaddr',
        'getnetbyname',
        'getnetent',
        'getpeername',
        'getpgrp',
        'getpriority',
        'getprotobyname',
        'getprotobynumber',
        'getprotoent',
        'getpwent',
        'getpwnam',
        'getpwuid',
        'getservbyname',
        'getservbyport',
        'getservent',
        'getsockname',
        'getsockopt',
        'given',
        'glob',
        'gmtime',
        'goto',
        'grep',
        'gt',
        'hex',
        'if',
        'index',
        'int',
        'ioctl',
        'join',
        'keys',
        'kill',
        'last',
        'lc',
        'lcfirst',
        'length',
        'link',
        'listen',
        'local',
        'localtime',
        'log',
        'lstat',
        'lt',
        'ma',
        'map',
        'mkdir',
        'msgctl',
        'msgget',
        'msgrcv',
        'msgsnd',
        'my',
        'ne',
        'next',
        'no',
        'not',
        'oct',
        'open',
        'opendir',
        'or',
        'ord',
        'our',
        'pack',
        'package',
        'pipe',
        'pop',
        'pos',
        'print',
        'printf',
        'prototype',
        'push',
        'q|0',
        'qq',
        'quotemeta',
        'qw',
        'qx',
        'rand',
        'read',
        'readdir',
        'readline',
        'readlink',
        'readpipe',
        'recv',
        'redo',
        'ref',
        'rename',
        'require',
        'reset',
        'return',
        'reverse',
        'rewinddir',
        'rindex',
        'rmdir',
        'say',
        'scalar',
        'seek',
        'seekdir',
        'select',
        'semctl',
        'semget',
        'semop',
        'send',
        'setgrent',
        'sethostent',
        'setnetent',
        'setpgrp',
        'setpriority',
        'setprotoent',
        'setpwent',
        'setservent',
        'setsockopt',
        'shift',
        'shmctl',
        'shmget',
        'shmread',
        'shmwrite',
        'shutdown',
        'sin',
        'sleep',
        'socket',
        'socketpair',
        'sort',
        'splice',
        'split',
        'sprintf',
        'sqrt',
        'srand',
        'stat',
        'state',
        'study',
        'sub',
        'substr',
        'symlink',
        'syscall',
        'sysopen',
        'sysread',
        'sysseek',
        'system',
        'syswrite',
        'tell',
        'telldir',
        'tie',
        'tied',
        'time',
        'times',
        'tr',
        'truncate',
        'uc',
        'ucfirst',
        'umask',
        'undef',
        'unless',
        'unlink',
        'unpack',
        'unshift',
        'untie',
        'until',
        'use',
        'utime',
        'values',
        'vec',
        'wait',
        'waitpid',
        'wantarray',
        'warn',
        'when',
        'while',
        'write',
        'x|0',
        'xor',
        'y|0',
    ];
    // https://perldoc.perl.org/perlre#Modifiers
    const REGEX_MODIFIERS = /[dualxmsipngr]{0,12}/; // aa and xx are valid, making max length 12
    const PERL_KEYWORDS = { $pattern: /[\w.]+/, keyword: KEYWORDS.join(' ') };
    const SUBST = {
        className: 'subst',
        begin: '[$@]\\{',
        end: '\\}',
        keywords: PERL_KEYWORDS,
    };
    const METHOD = { begin: /->\{/, end: /\}/ };
    const VAR = {
        variants: [
            { begin: /\$\d/ },
            { begin: concat(/[$%@](\^\w\b|#\w+(::\w+)*|\{\w+\}|\w+(::\w*)*)/, '(?![A-Za-z])(?![@$%])') },
            { begin: /[$%@][^\s\w{]/, relevance: 0 },
        ],
    };
    const STRING_CONTAINS = [BACKSLASH_ESCAPE, SUBST, VAR];
    const REGEX_DELIMS = [/!/, /\//, /\|/, /\?/, /'/, /"/, /#/];
    const PAIRED_DOUBLE_RE = (prefix, open, close = '\\1') => {
        const middle = (close === '\\1') ? close : concat(close, open);
        return concat(concat('(?:', prefix, ')'), open, /(?:\\.|[^\\/])*?/, middle, /(?:\\.|[^\\/])*?/, close, REGEX_MODIFIERS);
    };
    const PAIRED_RE = (prefix, open, close) => concat(concat('(?:', prefix, ')'), open, /(?:\\.|[^\\/])*?/, close, REGEX_MODIFIERS);
    const PERL_DEFAULT_CONTAINS = [
        VAR,
        HASH_COMMENT_MODE,
        COMMENT(/^=\w/, /=cut/, { endsWithParent: true }),
        METHOD,
        {
            className: 'string',
            contains: STRING_CONTAINS,
            variants: [
                { begin: 'q[qwxr]?\\s*\\(', end: '\\)', relevance: 5 },
                { begin: 'q[qwxr]?\\s*\\[', end: '\\]', relevance: 5 },
                { begin: 'q[qwxr]?\\s*\\{', end: '\\}', relevance: 5 },
                { begin: 'q[qwxr]?\\s*\\|', end: '\\|', relevance: 5 },
                { begin: 'q[qwxr]?\\s*<', end: '>', relevance: 5 },
                { begin: 'qw\\s+q', end: 'q', relevance: 5 },
                { begin: '\'', end: '\'', contains: [BACKSLASH_ESCAPE] },
                { begin: '"', end: '"' },
                { begin: '`', end: '`', contains: [BACKSLASH_ESCAPE] },
                { begin: /\{\w+\}/, relevance: 0 },
                { begin: '-?\\w+\\s*=>', relevance: 0 },
            ],
        },
        {
            className: 'number',
            begin: '(\\b0[0-7_]+)|(\\b0x[0-9a-fA-F_]+)|(\\b[1-9][0-9_]*(\\.[0-9_]+)?)|[0_]\\b',
            relevance: 0,
        },
        {
            begin: `(\\/\\/|${RE_STARTERS_RE}|\\b(split|return|print|reverse|grep)\\b)\\s*`,
            keywords: 'split return print reverse grep',
            relevance: 0,
            contains: [
                HASH_COMMENT_MODE,
                {
                    className: 'regexp',
                    variants: [
                        { begin: PAIRED_DOUBLE_RE('s|tr|y', either(...REGEX_DELIMS, { capture: true })) },
                        { begin: PAIRED_DOUBLE_RE('s|tr|y', '\\(', '\\)') },
                        { begin: PAIRED_DOUBLE_RE('s|tr|y', '\\[', '\\]') },
                        { begin: PAIRED_DOUBLE_RE('s|tr|y', '\\{', '\\}') },
                    ],
                    relevance: 2,
                },
                {
                    className: 'regexp',
                    variants: [
                        { begin: /(m|qr)\/\//, relevance: 0 },
                        { begin: PAIRED_RE('(?:m|qr)?', /\//, /\//) },
                        { begin: PAIRED_RE('m|qr', either(...REGEX_DELIMS, { capture: true }), /\1/) },
                        { begin: PAIRED_RE('m|qr', /\(/, /\)/) },
                        { begin: PAIRED_RE('m|qr', /\[/, /\]/) },
                        { begin: PAIRED_RE('m|qr', /\{/, /\}/) },
                    ],
                },
            ],
        },
        {
            className: 'function',
            beginKeywords: 'sub',
            end: '(\\s*\\(.*?\\))?[;{]',
            excludeEnd: true,
            relevance: 5,
            contains: [TITLE_MODE],
        },
        { begin: '-\\w\\b', relevance: 0 },
        {
            begin: '^__DATA__$',
            end: '^__END__$',
            subLanguage: 'mojolicious',
            contains: [{ begin: '^@@.*', end: '$', className: 'comment' }],
        },
    ];
    SUBST.contains = PERL_DEFAULT_CONTAINS;
    METHOD.contains = PERL_DEFAULT_CONTAINS;
    return {
        name: 'Perl',
        aliases: ['pl', 'pm'],
        keywords: PERL_KEYWORDS,
        contains: PERL_DEFAULT_CONTAINS,
    };
};

const pythonLanguage = () => {
    const IDENT_RE = /[\p{XID_Start}_]\p{XID_Continue}*/u;
    const RESERVED_WORDS = [
        'and',
        'as',
        'assert',
        'async',
        'await',
        'break',
        'class',
        'continue',
        'def',
        'del',
        'elif',
        'else',
        'except',
        'finally',
        'for',
        'from',
        'global',
        'if',
        'import',
        'in',
        'is',
        'lambda',
        'nonlocal|10',
        'not',
        'or',
        'pass',
        'raise',
        'return',
        'try',
        'while',
        'with',
        'yield',
    ];
    const BUILT_INS = [
        '__import__',
        'abs',
        'all',
        'any',
        'ascii',
        'bin',
        'bool',
        'breakpoint',
        'bytearray',
        'bytes',
        'callable',
        'chr',
        'classmethod',
        'compile',
        'complex',
        'delattr',
        'dict',
        'dir',
        'divmod',
        'enumerate',
        'eval',
        'exec',
        'filter',
        'float',
        'format',
        'frozenset',
        'getattr',
        'globals',
        'hasattr',
        'hash',
        'help',
        'hex',
        'id',
        'input',
        'int',
        'isinstance',
        'issubclass',
        'iter',
        'len',
        'list',
        'locals',
        'map',
        'max',
        'memoryview',
        'min',
        'next',
        'object',
        'oct',
        'open',
        'ord',
        'pow',
        'print',
        'property',
        'range',
        'repr',
        'reversed',
        'round',
        'set',
        'setattr',
        'slice',
        'sorted',
        'staticmethod',
        'str',
        'sum',
        'super',
        'tuple',
        'type',
        'vars',
        'zip',
    ];
    const LITERALS = [
        '__debug__',
        'Ellipsis',
        'False',
        'None',
        'NotImplemented',
        'True',
    ];
    // https://docs.python.org/3/library/typing.html
    // TODO: Could these be supplemented by a CamelCase matcher in certain
    // contexts, leaving these remaining only for relevance hinting?
    const TYPES = [
        'Any',
        'Callable',
        'Coroutine',
        'Dict',
        'List',
        'Literal',
        'Generic',
        'Optional',
        'Sequence',
        'Set',
        'Tuple',
        'Type',
        'Union',
    ];
    const KEYWORDS = {
        $pattern: /[A-Za-z]\w+|__\w+__/,
        keyword: RESERVED_WORDS,
        built_in: BUILT_INS,
        literal: LITERALS,
        type: TYPES,
    };
    const PROMPT = {
        className: 'meta',
        begin: /^(>>>|\.\.\.) /,
    };
    const SUBST = {
        className: 'subst',
        begin: /\{/,
        end: /\}/,
        keywords: KEYWORDS,
        illegal: /#/,
    };
    const LITERAL_BRACKET = {
        begin: /\{\{/,
        relevance: 0,
    };
    const STRING = {
        className: 'string',
        contains: [BACKSLASH_ESCAPE],
        variants: [
            {
                begin: /([uU]|[bB]|[rR]|[bB][rR]|[rR][bB])?'''/,
                end: /'''/,
                contains: [BACKSLASH_ESCAPE, PROMPT],
                relevance: 10,
            },
            {
                begin: /([uU]|[bB]|[rR]|[bB][rR]|[rR][bB])?"""/,
                end: /"""/,
                contains: [BACKSLASH_ESCAPE, PROMPT],
                relevance: 10,
            },
            {
                begin: /([fF][rR]|[rR][fF]|[fF])'''/,
                end: /'''/,
                contains: [BACKSLASH_ESCAPE, PROMPT, LITERAL_BRACKET, SUBST],
            },
            {
                begin: /([fF][rR]|[rR][fF]|[fF])"""/,
                end: /"""/,
                contains: [BACKSLASH_ESCAPE, PROMPT, LITERAL_BRACKET, SUBST],
            },
            {
                begin: /([uU]|[rR])'/,
                end: /'/,
                relevance: 10,
            },
            {
                begin: /([uU]|[rR])"/,
                end: /"/,
                relevance: 10,
            },
            {
                begin: /([bB]|[bB][rR]|[rR][bB])'/,
                end: /'/,
            },
            {
                begin: /([bB]|[bB][rR]|[rR][bB])"/,
                end: /"/,
            },
            {
                begin: /([fF][rR]|[rR][fF]|[fF])'/,
                end: /'/,
                contains: [BACKSLASH_ESCAPE, LITERAL_BRACKET, SUBST],
            },
            {
                begin: /([fF][rR]|[rR][fF]|[fF])"/,
                end: /"/,
                contains: [BACKSLASH_ESCAPE, LITERAL_BRACKET, SUBST],
            },
            APOS_STRING_MODE,
            QUOTE_STRING_MODE,
        ],
    };
    const digitPart = '[0-9](_?[0-9])*';
    const pointFloat = `(\\b(${digitPart}))?\\.(${digitPart})|\\b(${digitPart})\\.`;
    const _lookahead = `\\b|${RESERVED_WORDS.join('|')}`;
    const NUMBER = {
        className: 'number',
        relevance: 0,
        variants: [
            { begin: `(\\b(${digitPart})|(${pointFloat}))[eE][+-]?(${digitPart})[jJ]?(?=${_lookahead})` },
            { begin: `(${pointFloat})[jJ]?` },
            { begin: `\\b([1-9](_?[0-9])*|0+(_?0)*)[lLjJ]?(?=${_lookahead})` },
            { begin: `\\b0[bB](_?[01])+[lL]?(?=${_lookahead})` },
            { begin: `\\b0[oO](_?[0-7])+[lL]?(?=${_lookahead})` },
            { begin: `\\b0[xX](_?[0-9a-fA-F])+[lL]?(?=${_lookahead})` },
            { begin: `\\b(${digitPart})[jJ](?=${_lookahead})` },
        ],
    };
    const COMMENT_TYPE = {
        className: 'comment',
        begin: lookahead(/# type:/),
        end: /$/,
        keywords: KEYWORDS,
        contains: [
            { begin: /# type:/ },
            { begin: /#/, end: /\b\B/, endsWithParent: true },
        ],
    };
    const PARAMS = {
        className: 'params',
        variants: [
            { className: '', begin: /\(\s*\)/, skip: true },
            {
                begin: /\(/,
                end: /\)/,
                excludeBegin: true,
                excludeEnd: true,
                keywords: KEYWORDS,
                contains: ['self', PROMPT, NUMBER, STRING, HASH_COMMENT_MODE],
            },
        ],
    };
    SUBST.contains = [STRING, NUMBER, PROMPT];
    return {
        name: 'Python',
        aliases: ['py', 'gyp', 'ipython'],
        unicodeRegex: true,
        keywords: KEYWORDS,
        illegal: /(<\/|->|\?)|=>/,
        contains: [
            PROMPT,
            NUMBER,
            { begin: /\bself\b/ },
            { beginKeywords: 'if', relevance: 0 },
            STRING,
            COMMENT_TYPE,
            HASH_COMMENT_MODE,
            {
                match: [/\bdef/, /\s+/, IDENT_RE],
                scope: { 1: 'keyword', 3: 'title.function' },
                contains: [PARAMS],
            },
            {
                variants: [
                    { match: [/\bclass/, /\s+/, IDENT_RE, /\s*/, /\(\s*/, IDENT_RE, /\s*\)/] },
                    { match: [/\bclass/, /\s+/, IDENT_RE] },
                ],
                scope: {
                    1: 'keyword',
                    3: 'title.class',
                    6: 'title.class.inherited',
                },
            },
            {
                className: 'meta',
                begin: /^[\t ]*@/,
                end: /(?=#)|$/,
                contains: [NUMBER, STRING, PARAMS],
            },
        ],
    };
};

/*
  Language: TypeScript
  Author: Panu Horsmalahti <panu.horsmalahti@iki.fi>
  Contributors: Ike Ku <dempfi@yahoo.com>
  Description: TypeScript is a strict superset of JavaScript
  Website: https://www.typescriptlang.org
  Category: common, scripting
*/
const typescriptLanguage = () => {
    const tsLanguage = javascriptLanguage();
    const TYPES = [
        'any',
        'void',
        'number',
        'boolean',
        'string',
        'object',
        'never',
        'symbol',
        'bigint',
        'unknown',
    ];
    const NAMESPACE = {
        beginKeywords: 'namespace',
        end: /\{/,
        excludeEnd: true,
        contains: [tsLanguage.exports?.CLASS_REFERENCE],
    };
    const INTERFACE = {
        beginKeywords: 'interface',
        end: /\{/,
        excludeEnd: true,
        keywords: {
            keyword: 'interface extends',
            built_in: TYPES,
        },
        contains: [tsLanguage.exports?.CLASS_REFERENCE],
    };
    const USE_STRICT = {
        relevance: 10,
        className: 'meta',
        begin: /^\s*['"]use strict['"]/,
    };
    const TS_SPECIFIC_KEYWORDS = [
        'type',
        'namespace',
        'interface',
        'public',
        'private',
        'protected',
        'implements',
        'declare',
        'abstract',
        'readonly',
        'enum',
        'override',
    ];
    const KEYWORDS$1 = {
        $pattern: IDENT_RE,
        keyword: KEYWORDS.concat(TS_SPECIFIC_KEYWORDS),
        literal: LITERALS,
        built_in: BUILT_INS.concat(TYPES),
        'variable.language': BUILT_IN_VARIABLES,
    };
    const DECORATOR = { className: 'meta', begin: `@${IDENT_RE}` };
    const swapMode = (mode, label, replacement) => {
        const index = mode.contains.findIndex((m) => m.label === label);
        if (index === -1)
            throw new Error('Can not find mode to replace');
        mode.contains.splice(index, 1, replacement);
    };
    Object.assign(tsLanguage.keywords, KEYWORDS$1);
    if (tsLanguage.exports?.PARAMS_CONTAINS && Array.isArray(tsLanguage.exports.PARAMS_CONTAINS)) {
        tsLanguage.exports.PARAMS_CONTAINS.push(DECORATOR);
    }
    tsLanguage.contains = tsLanguage.contains.concat([DECORATOR, NAMESPACE, INTERFACE]);
    // TS gets a simpler shebang rule than JS
    swapMode(tsLanguage, 'shebang', SHEBANG());
    // JS use strict rule purposely excludes `asm` which makes no sense
    swapMode(tsLanguage, 'use_strict', USE_STRICT);
    const FUNCTION_DECLARATION = tsLanguage.contains.find((m) => m.label === 'func.def');
    FUNCTION_DECLARATION.relevance = 0; // () => {} is more typical in TypeScript
    Object.assign(tsLanguage, { name: 'TypeScript', aliases: ['ts', 'tsx'] });
    return tsLanguage;
};

const xmlLanguage = () => {
    const TAG_NAME_RE = concat(/[A-Z_]/, optional(/[A-Z0-9_.-]*:/), /[A-Z0-9_.-]*/);
    const XML_IDENT_RE = /[A-Za-z0-9._:-]+/;
    const XML_ENTITIES = { className: 'symbol', begin: /&[a-z]+;|&#[0-9]+;|&#x[a-f0-9]+;/ };
    const XML_META_KEYWORDS = {
        begin: /\s/,
        contains: [
            {
                className: 'keyword',
                begin: /#?[a-z_][a-z1-9_-]+/,
                illegal: /\n/,
            },
        ],
    };
    const XML_META_PAR_KEYWORDS = inherit(XML_META_KEYWORDS, {
        begin: /\(/,
        end: /\)/,
    });
    const APOS_META_STRING_MODE = inherit(APOS_STRING_MODE, { className: 'string' });
    const QUOTE_META_STRING_MODE = inherit(QUOTE_STRING_MODE, { className: 'string' });
    const TAG_INTERNALS = {
        endsWithParent: true,
        illegal: /</,
        relevance: 0,
        contains: [
            {
                className: 'attr',
                begin: XML_IDENT_RE,
                relevance: 0,
            },
            {
                begin: /=\s*/,
                relevance: 0,
                contains: [
                    {
                        className: 'string',
                        endsParent: true,
                        variants: [
                            {
                                begin: /"/,
                                end: /"/,
                                contains: [XML_ENTITIES],
                            },
                            {
                                begin: /'/,
                                end: /'/,
                                contains: [XML_ENTITIES],
                            },
                            { begin: /[^\s"'=<>`]+/ },
                        ],
                    },
                ],
            },
        ],
    };
    return {
        name: 'HTML, XML',
        aliases: [
            'html',
            'xhtml',
            'rss',
            'atom',
            'xjb',
            'xsd',
            'xsl',
            'plist',
            'wsf',
            'svg',
        ],
        caseInsensitive: true,
        contains: [
            {
                className: 'meta',
                begin: /<![a-z]/,
                end: />/,
                relevance: 10,
                contains: [
                    XML_META_KEYWORDS,
                    QUOTE_META_STRING_MODE,
                    APOS_META_STRING_MODE,
                    XML_META_PAR_KEYWORDS,
                    {
                        begin: /\[/,
                        end: /\]/,
                        contains: [
                            {
                                className: 'meta',
                                begin: /<![a-z]/,
                                end: />/,
                                contains: [
                                    XML_META_KEYWORDS,
                                    XML_META_PAR_KEYWORDS,
                                    QUOTE_META_STRING_MODE,
                                    APOS_META_STRING_MODE,
                                ],
                            },
                        ],
                    },
                ],
            },
            COMMENT(/<!--/, /-->/, { relevance: 10 }),
            { begin: /<!\[CDATA\[/, end: /\]\]>/, relevance: 10 },
            XML_ENTITIES,
            {
                className: 'meta',
                end: /\?>/,
                variants: [
                    { begin: /<\?xml/, relevance: 10, contains: [QUOTE_STRING_MODE] },
                    { begin: /<\?[a-z][a-z0-9]+/ },
                ],
            },
            {
                className: 'tag',
                begin: /<style(?=\s|>)/,
                end: />/,
                keywords: { name: 'style' },
                contains: [TAG_INTERNALS],
                starts: {
                    end: /<\/style>/,
                    returnEnd: true,
                    subLanguage: ['css', 'xml'],
                },
            },
            {
                className: 'tag',
                begin: /<script(?=\s|>)/,
                end: />/,
                keywords: { name: 'script' },
                contains: [TAG_INTERNALS],
                starts: {
                    end: /<\/script>/,
                    returnEnd: true,
                    subLanguage: ['javascript', 'handlebars', 'xml'],
                },
            },
            { className: 'tag', begin: /<>|<\/>/ },
            {
                className: 'tag',
                begin: concat(/</, lookahead(concat(TAG_NAME_RE, either(/\/>/, />/, /\s/)))),
                end: /\/?>/,
                contains: [{
                        className: 'name',
                        begin: TAG_NAME_RE,
                        relevance: 0,
                        starts: TAG_INTERNALS,
                    }],
            },
            {
                className: 'tag',
                begin: concat(/<\//, lookahead(concat(TAG_NAME_RE, />/))),
                contains: [
                    { className: 'name', begin: TAG_NAME_RE, relevance: 0 },
                    { begin: />/, relevance: 0, endsParent: true },
                ],
            },
        ],
    };
};

const nativeLanguages = {
    python: pythonLanguage,
    json: jsonLanguage,
    css: cssLanguage,
    perl: perlLanguage,
    javascript: javascriptLanguage,
    typescript: typescriptLanguage,
    xml: xmlLanguage,
};

const logger = {
    info: (message) => {
        console.info(message);
    },
    error: (message) => {
        console.error(message);
    },
    warn: (message) => {
        console.warn(message);
    },
};

/* eslint-disable @typescript-eslint/no-explicit-any */
const PLAINTEXT_LANGUAGE = { disableAutodetect: true, name: 'Plain Text', contains: [] };
const LANGUAGE_NOT_FOUND = "Could not find the language '{}', did you forget to load/include a language module?";
const MAX_KEYWORD_HITS = 7;
const NO_MATCH = Symbol('nomatch');
class NodeHighlight {
    _options;
    _languages;
    _safeMode;
    _aliases;
    _plugins;
    _logger;
    constructor(options, _logger = logger) {
        this._options = {
            escapeHTML: true,
            json: false,
            throwUnescapedHTML: false,
            noHighlightRe: /^(no-?highlight)$/i,
            languageDetectRe: /\blang(?:uage)?-([\w-]+)\b/i,
            classPrefix: 'nhgl-',
            languages: null,
            __emitter: TokenTreeEmitter,
            ...options,
        };
        this._languages = Object.create(null);
        this._aliases = Object.create(null);
        this.safeMode = true;
        this._logger = _logger;
        this._plugins = [];
        Object.keys(nativeLanguages).forEach((lang) => {
            this.registerLanguage(lang, nativeLanguages[lang]);
        });
    }
    get safeMode() {
        return this._safeMode;
    }
    set safeMode(state) {
        this._safeMode = state;
    }
    highlight = (code, languageName, ignoreIllegals = true) => {
        const fire = (event, args) => {
            this._plugins.forEach((plugin) => {
                if (event === 'before:highlight') {
                    plugin['before:highlight'](args);
                }
                else {
                    plugin['after:highlight'](args);
                }
            });
        };
        const context = {
            code,
            language: languageName,
        };
        fire('before:highlight', context);
        const result = context.result
            ? context.result
            : this._highlight(context.code, context.language, ignoreIllegals);
        result.code = context.code;
        fire('after:highlight', result);
        return result;
    };
    _highlight = (code, languageName, ignoreIllegals, continuation) => {
        // constants
        const keywordHits = Object.create(null);
        const language = this.getLanguage(languageName);
        if (!language) {
            logger.error(LANGUAGE_NOT_FOUND.replace('{}', languageName));
            throw new Error(`Unknown language: "${languageName}"`);
        }
        const md = compileLanguage(language);
        const continuations = {};
        const emitter = new this._options.__emitter(this._options);
        // variables
        let top = continuation || md;
        let modeBuffer = '';
        let relevance = 0;
        let resumeScanAtSamePosition = false;
        let iterations = 0;
        /**
         * Return keyword data if a match is a keyword
         * @param {CompiledMode} mode - current mode
         * @param {string} matchText - the textual match
         */
        const keywordData = (mode, matchText) => mode.keywords[matchText];
        const processKeywords = () => {
            if (!top.keywords) {
                emitter.addText(modeBuffer);
                return undefined;
            }
            let _lastIndex = 0;
            top.keywordPatternRe.lastIndex = 0;
            let match = top.keywordPatternRe.exec(modeBuffer);
            let buf = '';
            while (match) {
                buf += modeBuffer.substring(_lastIndex, match.index);
                const word = language.caseInsensitive ? match[0].toLowerCase() : match[0];
                const data = keywordData(top, word);
                if (data) {
                    const [kind, keywordRelevance] = data;
                    emitter.addText(buf);
                    buf = '';
                    keywordHits[word] = (keywordHits[word] || 0) + 1;
                    if (keywordHits[word] <= MAX_KEYWORD_HITS) {
                        relevance += keywordRelevance;
                    }
                    if (kind.startsWith('_')) {
                        buf += match[0];
                    }
                    else {
                        const cssClass = language.classNameAliases[kind] || kind;
                        emitter.addKeyword(match[0], cssClass);
                    }
                }
                else {
                    buf += match[0];
                }
                _lastIndex = top.keywordPatternRe.lastIndex;
                match = top.keywordPatternRe.exec(modeBuffer);
            }
            buf += modeBuffer.substring(_lastIndex);
            emitter.addText(buf);
            return undefined;
        };
        const processSubLanguage = () => {
            if (modeBuffer === '')
                return undefined;
            let _result = null;
            if (typeof top.subLanguage === 'string') {
                if (!this._languages[top.subLanguage]) {
                    emitter.addText(modeBuffer);
                    return undefined;
                }
                _result = this._highlight(modeBuffer, top.subLanguage, true, continuations[top.subLanguage]);
                continuations[top.subLanguage] = _result._top;
            }
            else {
                _result = this.highlightAuto(modeBuffer, top.subLanguage?.length ? top.subLanguage : null);
            }
            if (top.relevance > 0) {
                relevance += top.relevance;
            }
            emitter.addSubLanguage(_result._emitter, _result.language);
            return undefined;
        };
        const processBuffer = () => {
            if (top.subLanguage !== null && top.subLanguage !== undefined) {
                processSubLanguage();
            }
            else {
                processKeywords();
            }
            modeBuffer = '';
        };
        const emitMultiClass = (scope, match) => {
            let i = 1;
            const max = match.length - 1;
            while (i <= max) {
                if (scope._emit[i]) {
                    const klass = language.classNameAliases[scope[i]] || scope[i];
                    const text = match[i];
                    if (klass) {
                        emitter.addKeyword(text, klass);
                    }
                    else {
                        modeBuffer = text;
                        processKeywords();
                        modeBuffer = '';
                    }
                }
                i += 1;
            }
        };
        const startNewMode = (mode, match) => {
            if (mode.scope && typeof mode.scope === 'string') {
                emitter.openNode(language.classNameAliases[mode.scope] || mode.scope);
            }
            if (mode.beginScope) {
                if (mode.beginScope._wrap) {
                    emitter.addKeyword(modeBuffer, language.classNameAliases[mode.beginScope._wrap] || mode.beginScope._wrap);
                    modeBuffer = '';
                }
                else if (mode.beginScope._multi) {
                    emitMultiClass(mode.beginScope, match);
                    modeBuffer = '';
                }
            }
            top = Object.create(mode, { parent: { value: top } });
            return top;
        };
        const endOfMode = (mode, match, matchPlusRemainder) => {
            let matched = startsWith(mode.endRe, matchPlusRemainder);
            if (matched) {
                if (mode['on:end']) {
                    const resp = new Response(mode);
                    mode['on:end'](match, resp);
                    if (resp.isMatchIgnored) {
                        matched = false;
                    }
                }
                if (matched) {
                    while (mode.endsParent && mode.parent) {
                        mode = mode.parent;
                    }
                    return mode;
                }
            }
            if (mode.endsWithParent) {
                return endOfMode(mode.parent, match, matchPlusRemainder);
            }
            return undefined;
        };
        const doIgnore = (lexeme) => {
            if (top.matcher.regexIndex === 0) {
                // no more regexes to potentially match here, so we move the cursor forward one
                // space
                modeBuffer += lexeme[0];
                return 1;
            }
            // no need to move the cursor, we still have additional regexes to try and
            // match at this very spot
            resumeScanAtSamePosition = true;
            return 0;
        };
        const doBeginMatch = (match) => {
            const lexeme = match[0];
            const newMode = match.rule;
            const resp = new Response(newMode);
            const beforeCallbacks = [newMode.__beforeBegin, newMode['on:begin']];
            let _state = false;
            beforeCallbacks.forEach((cb) => {
                if (cb && !_state) {
                    cb(match, resp);
                    if (resp.isMatchIgnored) {
                        _state = true;
                    }
                }
            });
            if (_state) {
                return doIgnore(lexeme);
            }
            if (newMode.skip) {
                modeBuffer += lexeme;
            }
            else {
                if (newMode.excludeBegin) {
                    modeBuffer += lexeme;
                }
                processBuffer();
                if (!newMode.returnBegin && !newMode.excludeBegin) {
                    modeBuffer = lexeme;
                }
            }
            startNewMode(newMode, match);
            return newMode.returnBegin ? 0 : lexeme.length;
        };
        const doEndMatch = (match) => {
            const lexeme = match[0];
            const matchPlusRemainder = code.substring(match.index);
            const endMode = endOfMode(top, match, matchPlusRemainder);
            if (!endMode) {
                return NO_MATCH;
            }
            const origin = top;
            if (top.endScope && top.endScope._wrap) {
                processBuffer();
                emitter.addKeyword(lexeme, top.endScope._wrap);
            }
            else if (top.endScope && top.endScope._multi) {
                processBuffer();
                emitMultiClass(top.endScope, match);
            }
            else if (origin.skip) {
                modeBuffer += lexeme;
            }
            else {
                if (!(origin.returnEnd || origin.excludeEnd)) {
                    modeBuffer += lexeme;
                }
                processBuffer();
                if (origin.excludeEnd) {
                    modeBuffer = lexeme;
                }
            }
            do {
                if (top.scope) {
                    emitter.closeNode();
                }
                if (!top.skip && !top.subLanguage) {
                    relevance += top.relevance;
                }
                top = top.parent;
            } while (top !== endMode.parent);
            if (endMode.starts) {
                startNewMode(endMode.starts, match);
            }
            return origin.returnEnd ? 0 : lexeme.length;
        };
        const processContinuations = () => {
            const list = [];
            for (let current = top; current !== language; current = current.parent) {
                if (current.scope) {
                    list.unshift(current.scope);
                }
            }
            list.forEach((item) => emitter.openNode(item));
        };
        let lastMatch = {};
        const processLexeme = (textBeforeMatch, match) => {
            const lexeme = match && match[0];
            modeBuffer += textBeforeMatch;
            if (lexeme === null || lexeme === undefined) {
                processBuffer();
                return 0;
            }
            if (lastMatch.type === 'begin' && match.type === 'end' && lastMatch.index === match.index && lexeme === '') {
                modeBuffer += code.slice(match.index, match.index + 1);
                if (!this._safeMode) {
                    const err = new Error(`0 width match regex (${languageName})`);
                    err.languageName = languageName;
                    err.badRule = lastMatch.rule;
                    throw err;
                }
                return 1;
            }
            lastMatch = match;
            if (match.type === 'begin') {
                return doBeginMatch(match);
            }
            if (match.type === 'illegal' && !ignoreIllegals) {
                const err = new Error(`Illegal lexeme "${lexeme}" for mode "${(top.scope || '<unnamed>')}"`);
                err.mode = top;
                throw err;
            }
            else if (match.type === 'end') {
                const processed = doEndMatch(match);
                if (processed !== NO_MATCH) {
                    return processed;
                }
            }
            if (match.type === 'illegal' && lexeme === '')
                return 1;
            if (iterations > 100000 && iterations > match.index * 3) {
                const err = new Error('potential infinite loop, way more iterations than matches');
                throw err;
            }
            modeBuffer += lexeme;
            return lexeme.length;
        };
        let result = '';
        let index = 0;
        processContinuations();
        try {
            top.matcher.considerAll();
            for (;;) {
                iterations += 1;
                if (resumeScanAtSamePosition) {
                    resumeScanAtSamePosition = false;
                }
                else {
                    top.matcher.considerAll();
                }
                top.matcher.lastIndex = index;
                const match = top.matcher.exec(code);
                if (!match)
                    break;
                const beforeMatch = code.substring(index, match.index);
                const processedCount = processLexeme(beforeMatch, match);
                index = match.index + processedCount;
            }
            processLexeme(code.substring(index));
            emitter.closeAllNodes();
            emitter.finalize();
            result = this._options.json ? emitter.toJSON() : emitter.toHTML();
            return {
                language: languageName,
                value: result,
                relevance,
                illegal: false,
                _emitter: emitter,
                _top: top,
                _status: true,
            };
        }
        catch (err) {
            if (err.message && err.message.includes('Illegal')) {
                return {
                    language: languageName,
                    value: this._options.escapeHTML ? escapeHTML(code) : code,
                    illegal: true,
                    relevance: 0,
                    _illegalBy: {
                        message: err.message,
                        index,
                        context: code.slice(index - 100, index + 100),
                        mode: err.mode,
                        resultSoFar: result,
                    },
                    _emitter: emitter,
                    _status: false,
                };
            }
            if (this._safeMode) {
                return {
                    language: languageName,
                    value: this._options.escapeHTML ? escapeHTML(code) : code,
                    illegal: false,
                    relevance: 0,
                    errorRaised: err,
                    _emitter: emitter,
                    _top: top,
                    _status: false,
                };
            }
            throw err;
        }
    };
    /**
     * returns a valid highlight result, without actually doing any actual work,
     * auto highlight starts with this and it's possible for small snippets that
     * auto-detection may not find a better match
     * @param {string} code
     * @returns {Result}
     */
    justTextHighlightResult = (code) => {
        const result = {
            value: this._options.escapeHTML ? escapeHTML(code) : code,
            illegal: false,
            relevance: 0,
            _top: PLAINTEXT_LANGUAGE,
            _emitter: new this._options.__emitter(this._options),
            _status: true,
        };
        result._emitter.addText(code);
        return result;
    };
    /**
    Highlighting with language detection. Accepts a string with the code to
    highlight. Returns an object with the following properties:
  
    - language (detected language)
    - relevance (int)
    - value (an HTML string with highlighting markup)
    - secondBest (object with the same structure for second-best heuristically
      detected language, may be absent).
  
      @param {string} code
      @param {string[]} subset
      @returns {Result}
    */
    highlightAuto = (code, subset) => {
        const _languageSubset = subset || this._options.languages || Object.keys(this._languages);
        const plainText = this.justTextHighlightResult(code);
        const results = _languageSubset
            .filter(this.getLanguage)
            .filter(this.autoDetection)
            .map((name) => this._highlight(code, name, false));
        results.unshift(plainText);
        const sorted = results.sort((a, b) => {
            if (a.relevance !== b.relevance)
                return b.relevance - a.relevance;
            if (a.language && b.language) {
                if (this.getLanguage(a.language).supersetOf === b.language) {
                    return 1;
                }
                if (this.getLanguage(b.language).supersetOf === a.language) {
                    return -1;
                }
            }
            return 0;
        });
        const [best, secondBest] = sorted;
        const result = best;
        result.secondBest = secondBest;
        return result;
    };
    autoDetection = (name) => {
        const lang = this.getLanguage(name);
        return lang && !lang.disableAutodetect;
    };
    /**
     * Register a language grammar module
     *
     * @param {string} languageName
     * @param {LanguageDFn} langDFn
     */
    registerLanguage = (languageName, langDFn) => {
        let lang = null;
        try {
            lang = langDFn();
        }
        catch (err) {
            this._logger.error(`Language definition for ${languageName} could not be registered.`);
            if (!this._safeMode) {
                throw err;
            }
            else {
                logger.error(err);
            }
            lang = PLAINTEXT_LANGUAGE;
        }
        if (!lang.name) {
            lang.name = languageName;
        }
        this._languages[languageName] = lang;
        lang.rawDefinition = langDFn.bind(null);
        if (lang.aliases) {
            this.registerAliases(lang.aliases, languageName);
        }
    };
    /**
     * Register language aliases.
     *
     * @param {string|string[]} aliasList Single alias or list of aliases.
     * @param {string} languageName
     */
    registerAliases = (aliasList, languageName) => {
        const _aliasList = typeof aliasList === 'string' ? [aliasList] : [...aliasList];
        _aliasList.forEach((alias) => {
            this._aliases[alias.toLowerCase()] = languageName;
        });
    };
    /**
     * @returns {string[]} List of language internal names
     */
    listLanguages = () => Object.keys(this._languages);
    /**
     * @param {string} name - name of the language to retrieve
     * @returns {Language | undefined}
     */
    getLanguage = (name) => {
        const _name = (name || '').toLowerCase();
        return this._languages[_name] || this._languages[this._aliases[_name]];
    };
}

module.exports = NodeHighlight;
